import logging
import os

import cv2 as cv
import numpy as np

import sys
import getopt
from glob import glob

from stereoids import mkdir_p


def splitfn(fn):
    path, fn = os.path.split(fn)
    name, ext = os.path.splitext(fn)
    return path, name, ext


class Calibrator:
    def __init__(self, pattern_size=(7, 6), square_size=1.0, debug_dir=None):
        self.pattern_size = pattern_size
        self.square_size = square_size
        self.size = None

        self.pattern_points = np.zeros((np.prod(self.pattern_size), 3), np.float32)
        self.pattern_points[:, :2] = np.indices(self.pattern_size).T.reshape(-1, 2)
        self.pattern_points *= self.square_size

        self.debug_dir = debug_dir
        self.log = logging.getLogger(self.__class__.__name__)

    @property
    def h(self):
        return self.size[0]

    @property
    def w(self):
        return self.size[1]

    @property
    def size_wh(self):
        return self.size[::-1]

    def print_camera_values(self, mtx_l, postfix=""):
        fovx, fovy, focalLength, principalPoint, aspectRatio = \
            cv.calibrationMatrixValues(mtx_l, self.size_wh, self.w, self.h)

        self.log.debug("fov_x{}: {}\nfov_y{}: {}".format(postfix, fovx, postfix, fovy))

    def _calibrate_single_camera(self, chessboards):

        obj_points = []
        img_points = []

        for (corners, pattern_points) in [x for x in chessboards if x is not None]:
            img_points.append(corners)
            obj_points.append(pattern_points)

        rms, mtx_l, dstc_l, rvecs, tvecs = cv.calibrateCamera(obj_points, img_points,
                                                              self.size_wh, None, None)
        return img_points, obj_points, rms, mtx_l, dstc_l, rvecs, tvecs

    def _processImage(self, img, image_id=None):
        if img is None:
            return None

        assert self.w == img.shape[1] and self.h == img.shape[0], (
                "size: %d x %d ... " % (img.shape[1], img.shape[0]))
        found, corners = cv.findChessboardCorners(img, self.pattern_size)
        if found:
            term = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_COUNT, 30, 0.1)
            cv.cornerSubPix(img, corners, (5, 5), (-1, -1), term)

        if self.debug_dir and image_id:
            mkdir_p(self.debug_dir)
            vis = cv.cvtColor(img, cv.COLOR_GRAY2BGR)
            cv.drawChessboardCorners(vis, self.pattern_size, corners, found)
            outfile = os.path.join(self.debug_dir, image_id + '_chess.png')
            cv.imwrite(outfile, vis)

        if not found:
            self.log.debug('chessboard not found {}'.format(image_id))
            return None

        return (corners.reshape(-1, 2), self.pattern_points)

    def _stereo_calibrate(self, dstm_l, dstm_r, imgp_l, imgp_r, mtx_l, mtx_r, obj_points):
        flags = cv.CALIB_FIX_INTRINSIC
        T = np.zeros((3, 1), dtype=np.float64)
        R = np.eye(3, dtype=np.float64)
        retval, cameraMatrix1, distCoeffs1, cameraMatrix2, distCoeffs2, R, T, E, F = cv.stereoCalibrate(
            obj_points,
            imgp_l,
            imgp_r,
            mtx_l,
            dstm_l,
            mtx_r,
            dstm_r,
            self.size_wh,
            R,
            T,
            flags=flags
        )
        return E, F, R, T, retval

    def calibrate(self, double_frame_generator):

        count = 0;
        self.size = None
        chessboards = list()
        for img_left, img_right in double_frame_generator:
            if self.size is None:
                self.size = img_left.shape[:2]

            chssbrd_l = self._processImage(img_left, "l_{:02}".format(count))
            chssbrd_r = self._processImage(img_right, "r_{:02}".format(count))
            count += 1
            if chssbrd_r is not None and chssbrd_l is not None:
                chessboards.append((chssbrd_l, chssbrd_r))

        imgp_l, obj_points, rms_l, mtx_l, dstm_l, rvecs_l, tvecs_l = \
            self._calibrate_single_camera([l for l, r in chessboards])
        imgp_r, obj_points, rms_r, mtx_r, dstm_r, rvecs_r, tvecs_r = \
            self._calibrate_single_camera([r for l, r in chessboards])

        self.log.debug("rms_l:\n{}".format(rms_l))
        self.log.debug("rms_r:\n{}".format(rms_r))
        self.log.debug("mtx_l:\n{}".format(mtx_l))
        self.log.debug("mtx_r:\n{}".format(mtx_r))
        self.print_camera_values(mtx_l, postfix="_l")
        self.print_camera_values(mtx_r, postfix="_r")

        E, F, R, T, retval = self._stereo_calibrate(dstm_l, dstm_r, imgp_l, imgp_r, mtx_l, mtx_r,
                                                    obj_points)

        self.log.debug("\nretval: {}".format(retval))
        self.log.debug("R:\n{}".format(R))
        self.log.debug("T:\n{}".format(T))
        self.log.debug("E:\n{}".format(E))
        self.log.debug("F:\n{}".format(F))

        newImageSize = (int(self.w * 1), int(self.h * 1))
        R_l, R_r, P_l, P_r, Q, roi1, roi2 = \
            cv.stereoRectify(mtx_l, dstm_l,
                             mtx_r, dstm_l,
                             self.size_wh, R, T,
                             flags=cv.CALIB_ZERO_DISPARITY,
                             alpha=1,
                             newImageSize=newImageSize
                             )
        self.log.debug("R_l:\n{}".format(R_l))
        self.log.debug("R_r:\n{}".format(R_r))
        self.log.debug("P_l:\n{}".format(P_l))
        self.log.debug("P_r:\n{}".format(P_r))
        self.log.debug("Q:\n{}".format(Q))
        self.log.debug("roi1:\n{}".format(roi1))
        self.log.debug("roi2:\n{}".format(roi2))

        mapx_l, mapy_l = cv.initUndistortRectifyMap(mtx_l, dstm_l, R_l, P_l, newImageSize,
                                                    cv.CV_16SC2)
        mapx_r, mapy_r = cv.initUndistortRectifyMap(mtx_r, dstm_r, R_r, P_r, newImageSize,
                                                    cv.CV_16SC2)

        self.log.debug("mapx_l:\n{}".format(mapx_l))
        self.log.debug("mapy_l:\n{}".format(mapy_l))
        self.log.debug("mapx_r:\n{}".format(mapx_r))
        self.log.debug("mapy_r:\n{}".format(mapy_r))

        return mapx_l, mapy_l, mapx_r, mapy_r, Q


class Rectifier:

    def __init__(self, mapx_l, mapy_l, mapx_r, mapy_r):
        self.mapx_l = mapx_l
        self.mapy_l = mapy_l
        self.mapx_r = mapx_r
        self.mapy_r = mapy_r

    def rectify(self, img_l, img_r):
        stereo_undist_l = cv.remap(img_l, self.mapx_l, self.mapy_l, cv.INTER_LINEAR)
        stereo_undist_r = cv.remap(img_r, self.mapx_r, self.mapy_r, cv.INTER_LINEAR)

        return stereo_undist_l, stereo_undist_r


def file_double_image_generator(imgs_left, imgs_right):
    for fn_l, fn_r in zip(imgs_left, imgs_right):
        img_l = cv.cvtColor(cv.imread(fn_l), cv.COLOR_BGR2GRAY)
        img_r = cv.cvtColor(cv.imread(fn_r), cv.COLOR_BGR2GRAY)
        yield img_l, img_r


def main():
    args, img_names = getopt.getopt(sys.argv[1:], '', ['debug=', 'square_size=', 'threads='])

    args = dict(args)
    args.setdefault('--debug', './output/')
    args.setdefault('--square_size', 1)
    args.setdefault('--threads', 4)

    square_size = float(args.get('--square_size'))
    debug_dir = args.get("--debug")
    threads_num = int(args.get('--threads'))

    imgs_left = sorted(glob(img_names[0]))
    imgs_right = sorted(glob(img_names[1]))
    if debug_dir:
        logging.basicConfig(level=logging.DEBUG)

    gen = file_double_image_generator(imgs_left, imgs_right)

    calibrator = Calibrator(pattern_size=(6, 9), square_size=square_size, debug_dir=debug_dir)

    mapx_l, mapy_l, mapx_r, mapy_r, Q = calibrator.calibrate(gen)

    np.savetxt("q.txt", Q)
    np.save("q.npy", Q)
    np.save("mapx_l.npy", mapx_l)
    np.save("mapy_l.npy", mapy_l)
    np.save("mapx_r.npy", mapx_r)
    np.save("mapy_r.npy", mapy_r)

    rectifier = Rectifier(mapx_l, mapy_l, mapx_r, mapy_r)

    for img_l, img_r in file_double_image_generator(imgs_left, imgs_right):
        rect_l, rect_r = rectifier.rectify(img_l, img_r)
        cv.imshow("left_orig", img_l)
        cv.imshow("right_orig", img_r)
        cv.imshow("left", rect_l)
        cv.imshow("right", rect_r)
        cv.waitKey(0)


if __name__ == '__main__':
    print(__doc__)
    main()
    cv.destroyAllWindows()



import os

import cv2 as cv

from stereoids import mkdir_p


class Capture():

    def __init__(self, dev_l, dev_r, outdir="capture", prefix="", pattern_size=None, index_start=0):
        self.dev_l = dev_l
        self.dev_r = dev_r
        self.outdir = outdir
        self.prefix = prefix
        self.pattern_size = pattern_size
        self.index_start = index_start

    def start(self):

        cap_left = cv.VideoCapture(self.dev_l)
        cap_right = cv.VideoCapture(self.dev_r)

        capture_count = self.index_start
        while 1:
            ret_l, img_l = cap_left.read()
            ret_r, img_r = cap_right.read()

            if self.pattern_size:
                img_bw_l = cv.cvtColor(img_l, cv.COLOR_BGR2GRAY)
                img_bw_r = cv.cvtColor(img_r, cv.COLOR_BGR2GRAY)
                retval, corners = cv.findChessboardCorners(img_bw_l, patternSize=self.pattern_size)
                img_bw_l = cv.cvtColor(img_bw_l, cv.COLOR_GRAY2BGR)
                cv.drawChessboardCorners(img_bw_l, self.pattern_size, corners, retval)
                retval, corners = cv.findChessboardCorners(img_bw_r, patternSize=self.pattern_size)
                img_bw_r = cv.cvtColor(img_bw_r, cv.COLOR_GRAY2BGR)
                cv.drawChessboardCorners(img_bw_r, self.pattern_size, corners, retval)

                cv.imshow("left", img_bw_l)
                cv.imshow("right", img_bw_r)
            else:
                cv.imshow("left", img_l)
                cv.imshow("right", img_r)

            wait_key = cv.waitKey(1)
            key = wait_key & 0xFF
            if key == ord('q'):
                break
            if key == ord('t'):
                outfile_l = os.path.join(self.outdir, self.prefix + 'left_{:02}.png'.format(capture_count))
                outfile_r = os.path.join(self.outdir, self.prefix + 'right_{:02}.png'.format(capture_count))

                mkdir_p(self.outdir)
                cv.imwrite(outfile_l, img_l)
                cv.imwrite(outfile_r, img_r)
                capture_count += 1


def main():
    capture = Capture(0, 2, pattern_size=(6, 7), index_start=21)
    capture.start()


if __name__ == '__main__':
    print(__doc__)
    main()
    cv.destroyAllWindows()



import errno
import os


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



from corrector import ImgCorrector
import cv2 as cv


class TransformationCorrector(ImgCorrector):

    def __init__(self, mapx_l, mapy_l, mapx_r, mapy_r):
        self.mapx_l = mapx_l
        self.mapy_l = mapy_l
        self.mapx_r = mapx_r
        self.mapy_r = mapy_r

    def rectify(self, img_l, img_r):
        stereo_undist_l = cv.remap(img_l, self.mapx_l, self.mapy_l, cv.INTER_LINEAR)
        stereo_undist_r = cv.remap(img_r, self.mapx_r, self.mapy_r, cv.INTER_LINEAR)

        return stereo_undist_l, stereo_undist_r

    def corrected_generator(self, img_pair_generator):
        for img_l, img_r in img_pair_generator:
            yield self.rectify(img_l, img_r)

    def get_corrected_generator(self, img_pair_generator):
        return self.corrected_generator(img_pair_generator)



"""This module will identify markers of a given type and try to position them in 3d space"""
import argparse

import cv2 as cv
import numpy as np

from corrector import ImgCorrector
from corrector.transformation_corrector import TransformationCorrector
from detector import Detector
from detector.aruco import ArucoDetector


def find_unique_markers(marker_ids, corners):
    unique_markers = {}

    seen_ids = set()
    for i, marker_id in enumerate(marker_ids):

        int_marker_id = int(marker_id)
        if int_marker_id in seen_ids:
            try:
                unique_markers.pop(int_marker_id)
            except KeyError:
                pass
        else:
            seen_ids.add(int_marker_id)
            unique_markers[int_marker_id] = corners[i]
    return unique_markers


def find_marker_pairs(ids_l, corners_l, ids_r, corners_r):
    unique_l = find_unique_markers(ids_l, corners_l)
    unique_r = find_unique_markers(ids_r, corners_r)

    marker_pairs = {}
    for id_l in unique_l:
        if id_l in unique_r:
            marker_pairs[id_l] = (unique_l[id_l], unique_r[id_l])

    return marker_pairs


def get_position_from_corners(corners):

    return (corners[0][0] + corners[0][1] + corners[0][2] + corners[0][3]) / 4


def calculate_marker_disparity(marker_pairs):
    disparity_map = {}
    for marker_id in marker_pairs:
        corners_l, corners_r = marker_pairs[marker_id]
        pos_l = get_position_from_corners(corners_l)
        pos_r = get_position_from_corners(corners_r)

        disparity = pos_l[0] - pos_r[0];

        disparity_map[marker_id] = np.array([[[pos_l[0], pos_l[1], disparity]]])
    return disparity_map


class Locator:

    def __init__(self, Q, marker_detector: Detector, img_corrector: ImgCorrector = None):
        self.Q = Q
        self.marker_detector = marker_detector
        self.img_corrector = img_corrector

    def calculate_marker_positions(self, marker_disparity):

        marker_positions = {}
        for marker_id in marker_disparity:
            position = cv.perspectiveTransform(marker_disparity[marker_id], self.Q)
            marker_positions[marker_id] = position

        return marker_positions

    def locate(self, img_generator, callback):

        if self.img_corrector is not None:
            img_generator = self.img_corrector.get_corrected_generator(img_generator)

        for img_l, img_r in img_generator:
            ids_l, corners_l = self.marker_detector.detect(img_l)
            ids_r, corners_r = self.marker_detector.detect(img_r)

            if ids_l is not None and ids_r is not None:
                marker_pairs = find_marker_pairs(ids_l, corners_l, ids_r, corners_r)
                marker_disparity = calculate_marker_disparity(marker_pairs)
                marker_positions = self.calculate_marker_positions(marker_disparity)
                callback(marker_positions)


def main():
    parser = argparse.ArgumentParser(description='start a test marker location')

    parser.add_argument('--Q', type=str, default="q.npy", help='disparity-to-depth mapping matrix')
    parser.add_argument('--mapx_l', type=str, default="mapx_l.npy", help='transformation map x director left')
    parser.add_argument('--mapy_l', type=str, default="mapy_l.npy", help='transformation map y director left')
    parser.add_argument('--mapx_r', type=str, default="mapx_r.npy", help='transformation map x director right')
    parser.add_argument('--mapy_r', type=str, default="mapy_r.npy", help='transformation map y director right')
    args = parser.parse_args()

    # read correction parameter
    Q = np.load(args.Q)
    mapx_l = np.load(args.mapx_l)
    mapy_l = np.load(args.mapy_l)
    mapx_r = np.load(args.mapx_r)
    mapy_r = np.load(args.mapy_r)

    detector = ArucoDetector()
    corrector = TransformationCorrector(mapx_l, mapy_l, mapx_r, mapy_r)

    locator = Locator(Q, detector, corrector)

    def img_generator(l, r):
        cap_l = cv.VideoCapture(l)
        cap_r = cv.VideoCapture(r)

        while 1:
            ret_l, img_l = cap_l.read()
            ret_r, img_r = cap_r.read()

            cv.imshow("left", img_l)
            cv.imshow("right", img_r)

            wait_key = cv.waitKey(1)
            if wait_key != -1:
                break

        cv.destroyAllWindows()
        while True:
            ret_l, img_l = cap_l.read()
            ret_r, img_r = cap_r.read()
            img_l = cv.cvtColor(img_l, cv.COLOR_BGR2GRAY)
            img_r = cv.cvtColor(img_r, cv.COLOR_BGR2GRAY)
            yield img_l, img_r

    def callback(detections):

        for marker_id in detections:
            print(str(marker_id) + "\n" + str(detections[marker_id]))

    locator.locate(img_generator(2, 0), callback)


if __name__ == "__main__":
    main()

    cv.destroyAllWindows()



import cv2.aruco as aruco

from detector import Detector


class ArucoDetector(Detector):

    def __init__(self, dict=aruco.DICT_ARUCO_ORIGINAL):
        self.parameters = aruco.DetectorParameters_create()
        self.aruco_dict = aruco.Dictionary_get(dict)

    def detect(self, frame):
        corners, ids, _ = aruco.detectMarkers(frame, self.aruco_dict, parameters=self.parameters)

        return ids, corners



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
VERSION = open(os.path.join(here, 'VERSION.txt')).read()

REQUIREMENTS = [
    'opencv-contrib-python',
    'numpy',
]

setup(
    name='stereoids',
    version=VERSION,
    packages=find_packages(),
    url='https://github.com/sebastian-tschen/stereoids',
    license='MIT',
    author='Sebastian Behrens',
    author_email='sebastian.tschen.behrens@gmail.com',
    description='a simple library for 3d positioning markers via stereo vision',
    install_requires=REQUIREMENTS,
)
